@@ -83,10 +83,7 @@ static void validate_bpfpath_is_safe(const char *path)
 	}
 	// strtok_r() modifies its first argument, so work on a copy
 	char *tokenized SC_CLEANUP(sc_cleanup_string) = NULL;
-	tokenized = strdup(path);
-	if (tokenized == NULL) {
-		die("cannot allocate memory for copy of path");
-	}
+	tokenized = sc_strdup(path);
 	// allocate a string large enough to hold path, and initialize it to
 	// '/'
 	size_t checked_path_size = strlen(path) + 1;
@@ -110,10 +107,7 @@ static void validate_bpfpath_is_safe(const char *path)
 	char *buf_token = strtok_r(tokenized, "/", &buf_saveptr);
 	while (buf_token != NULL) {
 		char *prev SC_CLEANUP(sc_cleanup_string) = NULL;
-		prev = strdup(checked_path);	// needed by vsnprintf in sc_must_snprintf
-		if (prev == NULL) {
-			die("cannot allocate memory for copy of checked_path");
-		}
+		prev = sc_strdup(checked_path);	// needed by vsnprintf in sc_must_snprintf
 		// append '<buf_token>' if checked_path is '/', otherwise '/<buf_token>'
 		if (strlen(checked_path) == 1) {
 			sc_must_snprintf(checked_path, checked_path_size,
