@@ -19,9 +19,14 @@
 
 package builtin
 
-import (
-	"github.com/snapcore/snapd/interfaces"
-)
+const gsettingsSummary = `allows access to any gsettings item of current user`
+
+const gsettingsBaseDeclarationSlots = `
+  gsettings:
+    allow-installation:
+      slot-snap-type:
+        - core
+`
 
 const gsettingsConnectedPlugAppArmor = `
 # Description: Can access global gsettings of the user's session. Restricted
@@ -39,11 +44,13 @@ dbus (receive, send)
     peer=(label=unconfined),
 `
 
-// NewGsettingsInterface returns a new "gsettings" interface.
-func NewGsettingsInterface() interfaces.Interface {
-	return &commonInterface{
-		name: "gsettings",
+func init() {
+	registerIface(&commonInterface{
+		name:                  "gsettings",
+		summary:               gsettingsSummary,
+		implicitOnClassic:     true,
 		connectedPlugAppArmor: gsettingsConnectedPlugAppArmor,
+		baseDeclarationSlots:  gsettingsBaseDeclarationSlots,
 		reservedForOS:         true,
-	}
+	})
 }
