@@ -223,7 +223,7 @@ func (s *hookManagerSuite) TestHookTaskEnsure(c *C) {
 	defer s.state.Unlock()
 
 	c.Assert(s.context, NotNil, Commentf("Expected handler generator to be called with a valid context"))
-	c.Check(s.context.SnapName(), Equals, "test-snap")
+	c.Check(s.context.InstanceName(), Equals, "test-snap")
 	c.Check(s.context.SnapRevision(), Equals, snap.R(1))
 	c.Check(s.context.HookName(), Equals, "configure")
 
@@ -302,7 +302,7 @@ func (s *hookManagerSuite) TestHookHijackingHappy(c *C) {
 	c.Check(s.command.Calls(), HasLen, 0)
 
 	c.Assert(s.context, NotNil)
-	c.Check(s.context.SnapName(), Equals, "test-snap")
+	c.Check(s.context.InstanceName(), Equals, "test-snap")
 	c.Check(s.context.SnapRevision(), Equals, snap.R(1))
 	c.Check(s.context.HookName(), Equals, "configure")
 
@@ -329,7 +329,7 @@ func (s *hookManagerSuite) TestHookHijackingUnHappy(c *C) {
 	c.Check(s.command.Calls(), HasLen, 0)
 
 	c.Assert(s.context, NotNil)
-	c.Check(s.context.SnapName(), Equals, "test-snap")
+	c.Check(s.context.InstanceName(), Equals, "test-snap")
 	c.Check(s.context.SnapRevision(), Equals, snap.R(1))
 	c.Check(s.context.HookName(), Equals, "configure")
 
@@ -1013,13 +1013,13 @@ func (s *hookManagerSuite) TestHookTasksForDifferentSnapsRunConcurrently(c *C) {
 		testSnap2HookCalls++
 	})
 	s.manager.Register(regexp.MustCompile("prepare-device"), func(context *hookstate.Context) hookstate.Handler {
-		if context.SnapName() == "test-snap-1" {
+		if context.InstanceName() == "test-snap-1" {
 			return mockHandler1
 		}
-		if context.SnapName() == "test-snap-2" {
+		if context.InstanceName() == "test-snap-2" {
 			return mockHandler2
 		}
-		c.Fatalf("unknown snap: %s", context.SnapName())
+		c.Fatalf("unknown snap: %s", context.InstanceName())
 		return nil
 	})
 
