@@ -29,6 +29,7 @@ import (
 	"strings"
 
 	. "gopkg.in/check.v1"
+	"gopkg.in/yaml.v2"
 
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/snap"
@@ -185,6 +186,10 @@ apps:
    command: bar
  sample:
    command: foobar
+   command-chain: [chain]
+hooks:
+ configure:
+  command-chain: [hookchain]
 `
 
 func (s *infoSuite) TestReadInfo(c *C) {
@@ -200,6 +205,8 @@ func (s *infoSuite) TestReadInfo(c *C) {
 	c.Check(snapInfo2.Summary(), Equals, "esummary")
 
 	c.Check(snapInfo2.Apps["app"].Command, Equals, "foo")
+	c.Check(snapInfo2.Apps["sample"].CommandChain, DeepEquals, []string{"chain"})
+	c.Check(snapInfo2.Hooks["configure"].CommandChain, DeepEquals, []string{"hookchain"})
 
 	c.Check(snapInfo2, DeepEquals, snapInfo1)
 }
@@ -218,6 +225,8 @@ func (s *infoSuite) TestReadInfoWithInstance(c *C) {
 	c.Check(snapInfo2.Summary(), Equals, "instance summary")
 
 	c.Check(snapInfo2.Apps["app"].Command, Equals, "foo")
+	c.Check(snapInfo2.Apps["sample"].CommandChain, DeepEquals, []string{"chain"})
+	c.Check(snapInfo2.Hooks["configure"].CommandChain, DeepEquals, []string{"hookchain"})
 
 	c.Check(snapInfo2, DeepEquals, snapInfo1)
 }
@@ -325,8 +334,30 @@ func (s *infoSuite) TestReadInfoUnfindable(c *C) {
 	c.Check(info, IsNil)
 }
 
+func (s *infoSuite) TestReadInfoDanglingSymlink(c *C) {
+	si := &snap.SideInfo{Revision: snap.R(42), EditedSummary: "esummary"}
+	mpi := snap.MinimalPlaceInfo("sample", si.Revision)
+	p := filepath.Join(mpi.MountDir(), "meta", "snap.yaml")
+	c.Assert(os.MkdirAll(filepath.Dir(p), 0755), IsNil)
+	c.Assert(ioutil.WriteFile(p, []byte(`name: test`), 0644), IsNil)
+	c.Assert(os.MkdirAll(filepath.Dir(mpi.MountFile()), 0755), IsNil)
+	c.Assert(os.Symlink("/dangling", mpi.MountFile()), IsNil)
+
+	info, err := snap.ReadInfo("sample", si)
+	c.Check(err, IsNil)
+	c.Check(info.SnapName(), Equals, "test")
+	c.Check(info.Revision, Equals, snap.R(42))
+	c.Check(info.Summary(), Equals, "esummary")
+	c.Check(info.Size, Equals, int64(0))
+}
+
 // makeTestSnap here can also be used to produce broken snaps (differently from snaptest.MakeTestSnapWithFiles)!
-func makeTestSnap(c *C, yaml string) string {
+func makeTestSnap(c *C, snapYaml string) string {
+	var m struct {
+		Type string `yaml:"type"`
+	}
+	yaml.Unmarshal([]byte(snapYaml), &m) // yes, ignore the error
+
 	tmp := c.MkDir()
 	snapSource := filepath.Join(tmp, "snapsrc")
 
@@ -334,12 +365,12 @@ func makeTestSnap(c *C, yaml string) string {
 	c.Assert(err, IsNil)
 
 	// our regular snap.yaml
-	err = ioutil.WriteFile(filepath.Join(snapSource, "meta", "snap.yaml"), []byte(yaml), 0644)
+	err = ioutil.WriteFile(filepath.Join(snapSource, "meta", "snap.yaml"), []byte(snapYaml), 0644)
 	c.Assert(err, IsNil)
 
 	dest := filepath.Join(tmp, "foo.snap")
 	snap := squashfs.New(dest)
-	err = snap.Build(snapSource)
+	err = snap.Build(snapSource, m.Type)
 	c.Assert(err, IsNil)
 
 	return dest
@@ -447,7 +478,7 @@ type: app`
 	c.Assert(err, IsNil)
 
 	_, err = snap.ReadInfoFromSnapFile(snapf, nil)
-	c.Assert(err, ErrorMatches, "invalid snap name.*")
+	c.Assert(err, ErrorMatches, `invalid snap name.*`)
 }
 
 func (s *infoSuite) TestReadInfoFromSnapFileCatchesInvalidType(c *C) {
@@ -680,7 +711,7 @@ version: 1.0`
 		// Verify that the `test-hook` hook has now been loaded, and that it has
 		// no associated plugs.
 		c.Check(info.Hooks, HasLen, 1)
-		verifyImplicitHook(c, info, "test-hook")
+		verifyImplicitHook(c, info, "test-hook", nil)
 	})
 }
 
@@ -691,8 +722,8 @@ version: 1.0`
 		// Verify that both hooks have now been loaded, and that neither have any
 		// associated plugs.
 		c.Check(info.Hooks, HasLen, 2)
-		verifyImplicitHook(c, info, "foo")
-		verifyImplicitHook(c, info, "bar")
+		verifyImplicitHook(c, info, "foo", nil)
+		verifyImplicitHook(c, info, "bar", nil)
 	})
 }
 
@@ -705,7 +736,7 @@ version: 1.0`
 	s.checkInstalledSnapAndSnapFile(c, yaml, "SNAP", []string{"foo", "bar"}, func(c *C, info *snap.Info) {
 		// Verify that only foo has been loaded, not bar
 		c.Check(info.Hooks, HasLen, 1)
-		verifyImplicitHook(c, info, "foo")
+		verifyImplicitHook(c, info, "foo", nil)
 	})
 }
 
@@ -721,16 +752,125 @@ hooks:
 		// no associated plugs. Also verify that the `explicit` hook is still
 		// valid.
 		c.Check(info.Hooks, HasLen, 2)
-		verifyImplicitHook(c, info, "implicit")
+		verifyImplicitHook(c, info, "implicit", nil)
 		verifyExplicitHook(c, info, "explicit", []string{"test-plug"}, []string{"test-slot"})
 	})
 }
 
-func verifyImplicitHook(c *C, info *snap.Info, hookName string) {
+func (s *infoSuite) TestReadInfoExplicitHooks(c *C) {
+	yaml := `name: foo
+version: 1.0
+plugs:
+  test-plug:
+slots:
+  test-slot:
+hooks:
+  explicit:
+`
+	s.checkInstalledSnapAndSnapFile(c, yaml, "SNAP", []string{"explicit"}, func(c *C, info *snap.Info) {
+		c.Check(info.Hooks, HasLen, 1)
+		verifyExplicitHook(c, info, "explicit", []string{"test-plug"}, []string{"test-slot"})
+	})
+}
+
+func (s *infoSuite) TestReadInfoImplicitHookWithTopLevelPlugSlots(c *C) {
+	yaml := `name: foo
+version: 1.0
+plugs:
+  test-plug:
+slots:
+  test-slot:
+hooks:
+  explicit:
+    plugs: [test-plug,other-plug]
+    slots: [test-slot,other-slot]
+`
+
+	yaml2 := `name: foo
+version: 1.0
+plugs:
+  test-plug:
+slots:
+  test-slot:
+`
+	s.checkInstalledSnapAndSnapFile(c, yaml, "SNAP", []string{"implicit"}, func(c *C, info *snap.Info) {
+		c.Check(info.Hooks, HasLen, 2)
+		implicitHook := info.Hooks["implicit"]
+		c.Assert(implicitHook, NotNil)
+		c.Assert(implicitHook.Explicit, Equals, false)
+		c.Assert(implicitHook.Plugs, HasLen, 1)
+		c.Assert(implicitHook.Slots, HasLen, 1)
+
+		c.Check(info.Plugs, HasLen, 2)
+		c.Check(info.Slots, HasLen, 2)
+
+		plug := info.Plugs["test-plug"]
+		c.Assert(plug, NotNil)
+		c.Assert(implicitHook.Plugs["test-plug"], DeepEquals, plug)
+
+		slot := info.Slots["test-slot"]
+		c.Assert(slot, NotNil)
+		c.Assert(implicitHook.Slots["test-slot"], DeepEquals, slot)
+
+		explicitHook := info.Hooks["explicit"]
+		c.Assert(explicitHook, NotNil)
+		c.Assert(explicitHook.Explicit, Equals, true)
+		c.Assert(explicitHook.Plugs, HasLen, 2)
+		c.Assert(explicitHook.Slots, HasLen, 2)
+
+		plug = info.Plugs["test-plug"]
+		c.Assert(plug, NotNil)
+		c.Assert(explicitHook.Plugs["test-plug"], DeepEquals, plug)
+
+		slot = info.Slots["test-slot"]
+		c.Assert(slot, NotNil)
+		c.Assert(explicitHook.Slots["test-slot"], DeepEquals, slot)
+	})
+
+	s.checkInstalledSnapAndSnapFile(c, yaml2, "SNAP", []string{"implicit"}, func(c *C, info *snap.Info) {
+		c.Check(info.Hooks, HasLen, 1)
+		implicitHook := info.Hooks["implicit"]
+		c.Assert(implicitHook, NotNil)
+		c.Assert(implicitHook.Explicit, Equals, false)
+		c.Assert(implicitHook.Plugs, HasLen, 1)
+		c.Assert(implicitHook.Slots, HasLen, 1)
+
+		c.Check(info.Plugs, HasLen, 1)
+		c.Check(info.Slots, HasLen, 1)
+
+		plug := info.Plugs["test-plug"]
+		c.Assert(plug, NotNil)
+		c.Assert(implicitHook.Plugs["test-plug"], DeepEquals, plug)
+
+		slot := info.Slots["test-slot"]
+		c.Assert(slot, NotNil)
+		c.Assert(implicitHook.Slots["test-slot"], DeepEquals, slot)
+	})
+
+}
+
+func verifyImplicitHook(c *C, info *snap.Info, hookName string, plugNames []string) {
 	hook := info.Hooks[hookName]
 	c.Assert(hook, NotNil, Commentf("Expected hooks to contain %q", hookName))
 	c.Check(hook.Name, Equals, hookName)
-	c.Check(hook.Plugs, IsNil)
+
+	if len(plugNames) == 0 {
+		c.Check(hook.Plugs, IsNil)
+	}
+
+	for _, plugName := range plugNames {
+		// Verify that the HookInfo and PlugInfo point to each other
+		plug := hook.Plugs[plugName]
+		c.Assert(plug, NotNil, Commentf("Expected hook plugs to contain %q", plugName))
+		c.Check(plug.Name, Equals, plugName)
+		c.Check(plug.Hooks, HasLen, 1)
+		hook = plug.Hooks[hookName]
+		c.Assert(hook, NotNil, Commentf("Expected plug to be associated with hook %q", hookName))
+		c.Check(hook.Name, Equals, hookName)
+
+		// Verify also that the hook plug made it into info.Plugs
+		c.Check(info.Plugs[plugName], DeepEquals, plug)
+	}
 }
 
 func verifyExplicitHook(c *C, info *snap.Info, hookName string, plugNames []string, slotNames []string) {
@@ -1166,6 +1306,17 @@ func (s *infoSuite) TestExpandSnapVariables(c *C) {
 	c.Assert(info.ExpandSnapVariables("$SNAP_DATA/stuff"), Equals, "/var/snap/foo/42/stuff")
 	c.Assert(info.ExpandSnapVariables("$SNAP_COMMON/stuff"), Equals, "/var/snap/foo/common/stuff")
 	c.Assert(info.ExpandSnapVariables("$GARBAGE/rocks"), Equals, "/rocks")
+
+	info.InstanceKey = "instance"
+	// Despite setting the instance key the variables expand to the same
+	// value as before. This is because they are used from inside the mount
+	// namespace of the instantiated snap where the mount backend will
+	// ensure that the regular (non-instance) paths contain
+	// instance-specific code and data.
+	c.Assert(info.ExpandSnapVariables("$SNAP/stuff"), Equals, "/snap/foo/42/stuff")
+	c.Assert(info.ExpandSnapVariables("$SNAP_DATA/stuff"), Equals, "/var/snap/foo/42/stuff")
+	c.Assert(info.ExpandSnapVariables("$SNAP_COMMON/stuff"), Equals, "/var/snap/foo/common/stuff")
+	c.Assert(info.ExpandSnapVariables("$GARBAGE/rocks"), Equals, "/rocks")
 }
 
 func (s *infoSuite) TestStopModeTypeKillMode(c *C) {
@@ -1281,6 +1432,7 @@ func (s *infoSuite) TestDirAndFileHelpers(c *C) {
 	c.Check(snap.UserDataDir("/home/bob", "name", snap.R(1)), Equals, "/home/bob/snap/name/1")
 	c.Check(snap.UserCommonDataDir("/home/bob", "name"), Equals, "/home/bob/snap/name/common")
 	c.Check(snap.UserXdgRuntimeDir(12345, "name"), Equals, "/run/user/12345/snap.name")
+	c.Check(snap.UserSnapDir("/home/bob", "name"), Equals, "/home/bob/snap/name")
 
 	c.Check(snap.MountDir("name_instance", snap.R(1)), Equals, fmt.Sprintf("%s/name_instance/1", dirs.SnapMountDir))
 	c.Check(snap.MountFile("name_instance", snap.R(1)), Equals, "/var/lib/snapd/snaps/name_instance_1.snap")
@@ -1290,4 +1442,158 @@ func (s *infoSuite) TestDirAndFileHelpers(c *C) {
 	c.Check(snap.UserDataDir("/home/bob", "name_instance", snap.R(1)), Equals, "/home/bob/snap/name_instance/1")
 	c.Check(snap.UserCommonDataDir("/home/bob", "name_instance"), Equals, "/home/bob/snap/name_instance/common")
 	c.Check(snap.UserXdgRuntimeDir(12345, "name_instance"), Equals, "/run/user/12345/snap.name_instance")
+	c.Check(snap.UserSnapDir("/home/bob", "name_instance"), Equals, "/home/bob/snap/name_instance")
+}
+
+func (s *infoSuite) TestSortByType(c *C) {
+	infos := []*snap.Info{
+		{SuggestedName: "app1", Type: "app"},
+		{SuggestedName: "os1", Type: "os"},
+		{SuggestedName: "base1", Type: "base"},
+		{SuggestedName: "gadget1", Type: "gadget"},
+		{SuggestedName: "kernel1", Type: "kernel"},
+		{SuggestedName: "app2", Type: "app"},
+		{SuggestedName: "os2", Type: "os"},
+		{SuggestedName: "snapd", Type: "snapd"},
+		{SuggestedName: "base2", Type: "base"},
+		{SuggestedName: "gadget2", Type: "gadget"},
+		{SuggestedName: "kernel2", Type: "kernel"},
+	}
+	sort.Stable(snap.ByType(infos))
+
+	c.Check(infos, DeepEquals, []*snap.Info{
+		{SuggestedName: "snapd", Type: "snapd"},
+		{SuggestedName: "os1", Type: "os"},
+		{SuggestedName: "os2", Type: "os"},
+		{SuggestedName: "kernel1", Type: "kernel"},
+		{SuggestedName: "kernel2", Type: "kernel"},
+		{SuggestedName: "base1", Type: "base"},
+		{SuggestedName: "base2", Type: "base"},
+		{SuggestedName: "gadget1", Type: "gadget"},
+		{SuggestedName: "gadget2", Type: "gadget"},
+		{SuggestedName: "app1", Type: "app"},
+		{SuggestedName: "app2", Type: "app"},
+	})
+}
+
+func (s *infoSuite) TestSortByTypeAgain(c *C) {
+	core := &snap.Info{Type: snap.TypeOS}
+	base := &snap.Info{Type: snap.TypeBase}
+	app := &snap.Info{Type: snap.TypeApp}
+	snapd := &snap.Info{}
+	snapd.SideInfo = snap.SideInfo{RealName: "snapd"}
+
+	byType := func(snaps ...*snap.Info) []*snap.Info {
+		sort.Stable(snap.ByType(snaps))
+		return snaps
+	}
+
+	c.Check(byType(base, core), DeepEquals, []*snap.Info{core, base})
+	c.Check(byType(app, core), DeepEquals, []*snap.Info{core, app})
+	c.Check(byType(app, base), DeepEquals, []*snap.Info{base, app})
+	c.Check(byType(app, base, core), DeepEquals, []*snap.Info{core, base, app})
+	c.Check(byType(app, core, base), DeepEquals, []*snap.Info{core, base, app})
+
+	c.Check(byType(app, core, base, snapd), DeepEquals, []*snap.Info{snapd, core, base, app})
+	c.Check(byType(app, snapd, core, base), DeepEquals, []*snap.Info{snapd, core, base, app})
+}
+
+func (s *infoSuite) TestMedia(c *C) {
+	c.Check(snap.MediaInfos{}.Screenshots(), HasLen, 0)
+	c.Check(snap.MediaInfos{}.IconURL(), Equals, "")
+
+	media := snap.MediaInfos{
+		{
+			Type: "screenshot",
+			URL:  "https://example.com/shot1.svg",
+		}, {
+			Type: "icon",
+			URL:  "https://example.com/icon.png",
+		}, {
+			Type:   "screenshot",
+			URL:    "https://example.com/shot2.svg",
+			Width:  42,
+			Height: 17,
+		},
+	}
+
+	c.Check(media.IconURL(), Equals, "https://example.com/icon.png")
+	c.Check(media.Screenshots(), DeepEquals, []snap.ScreenshotInfo{
+		{
+			URL:  "https://example.com/shot1.svg",
+			Note: snap.ScreenshotsDeprecationNotice,
+		}, {
+			URL:    "https://example.com/shot2.svg",
+			Width:  42,
+			Height: 17,
+			Note:   snap.ScreenshotsDeprecationNotice,
+		},
+	})
+}
+
+func (s *infoSuite) TestSortApps(c *C) {
+	tcs := []struct {
+		err    string
+		apps   []*snap.AppInfo
+		sorted []string
+	}{{
+		apps: []*snap.AppInfo{
+			{Name: "bar", Before: []string{"baz"}},
+			{Name: "baz", After: []string{"bar", "foo"}},
+			{Name: "foo"},
+		},
+		sorted: []string{"bar", "foo", "baz"},
+	}, {
+		apps: []*snap.AppInfo{
+			{Name: "foo", After: []string{"bar", "zed"}},
+			{Name: "bar", Before: []string{"foo"}},
+			{Name: "baz", After: []string{"foo"}},
+			{Name: "zed"},
+		},
+		sorted: []string{"bar", "zed", "foo", "baz"},
+	}, {
+		apps: []*snap.AppInfo{
+			{Name: "foo", After: []string{"baz"}},
+			{Name: "bar", Before: []string{"baz"}},
+			{Name: "baz"},
+			{Name: "zed", After: []string{"foo", "bar", "baz"}},
+		},
+		sorted: []string{"bar", "baz", "foo", "zed"},
+	}, {
+		apps: []*snap.AppInfo{
+			{Name: "foo", Before: []string{"bar"}, After: []string{"zed"}},
+			{Name: "bar", Before: []string{"baz"}},
+			{Name: "baz", Before: []string{"zed"}},
+			{Name: "zed"},
+		},
+		err: `applications are part of a before/after cycle: ((foo|bar|baz|zed)(, )?){4}`,
+	}, {
+		apps: []*snap.AppInfo{
+			{Name: "foo", Before: []string{"bar"}},
+			{Name: "bar", Before: []string{"foo"}},
+			{Name: "baz", Before: []string{"foo"}, After: []string{"bar"}},
+		},
+		err: `applications are part of a before/after cycle: ((foo|bar|baz)(, )?){3}`,
+	}, {
+		apps: []*snap.AppInfo{
+			{Name: "baz", After: []string{"bar"}},
+			{Name: "foo"},
+			{Name: "bar", After: []string{"foo"}},
+		},
+		sorted: []string{"foo", "bar", "baz"},
+	}}
+	for _, tc := range tcs {
+		sorted, err := snap.SortServices(tc.apps)
+		if tc.err != "" {
+			c.Assert(err, ErrorMatches, tc.err)
+		} else {
+			c.Assert(err, IsNil)
+			c.Assert(sorted, HasLen, len(tc.sorted))
+			sortedNames := make([]string, len(sorted))
+			for i, app := range sorted {
+				sortedNames[i] = app.Name
+			}
+			c.Assert(sortedNames, DeepEquals, tc.sorted)
+		}
+	}
 }
