@@ -17,7 +17,7 @@
  *
  */
 
-package squashfs
+package squashfs_test
 
 import (
 	"errors"
@@ -31,10 +31,12 @@ import (
 	"time"
 
 	. "gopkg.in/check.v1"
+	"gopkg.in/yaml.v2"
 
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/snap/snapdir"
+	"github.com/snapcore/snapd/snap/squashfs"
 	"github.com/snapcore/snapd/testutil"
 )
 
@@ -47,12 +49,12 @@ type SquashfsTestSuite struct {
 
 var _ = Suite(&SquashfsTestSuite{})
 
-func makeSnap(c *C, manifest, data string) *Snap {
+func makeSnap(c *C, manifest, data string) *squashfs.Snap {
 	cur, _ := os.Getwd()
 	return makeSnapInDir(c, cur, manifest, data)
 }
 
-func makeSnapInDir(c *C, dir, manifest, data string) *Snap {
+func makeSnapContents(c *C, manifest, data string) string {
 	tmp := c.MkDir()
 	err := os.MkdirAll(filepath.Join(tmp, "meta", "hooks", "dir"), 0755)
 	c.Assert(err, IsNil)
@@ -79,9 +81,22 @@ func makeSnapInDir(c *C, dir, manifest, data string) *Snap {
 	err = ioutil.WriteFile(filepath.Join(tmp, "data.bin"), []byte(data), 0644)
 	c.Assert(err, IsNil)
 
+	return tmp
+}
+
+func makeSnapInDir(c *C, dir, manifest, data string) *squashfs.Snap {
+	snapType := "app"
+	var m struct {
+		Type string `yaml:"type"`
+	}
+	if err := yaml.Unmarshal([]byte(manifest), &m); err == nil && m.Type != "" {
+		snapType = m.Type
+	}
+
+	tmp := makeSnapContents(c, manifest, data)
 	// build it
-	snap := New(filepath.Join(dir, "foo.snap"))
-	err = snap.Build(tmp)
+	snap := squashfs.New(filepath.Join(dir, "foo.snap"))
+	err := snap.Build(tmp, snapType)
 	c.Assert(err, IsNil)
 
 	return snap
@@ -118,7 +133,7 @@ exec /bin/cp "$@"
 	defer cmd.Restore()
 	// mock link but still link
 	linked := 0
-	r := mockLink(func(a, b string) error {
+	r := squashfs.MockLink(func(a, b string) error {
 		linked++
 		return os.Link(a, b)
 	})
@@ -134,16 +149,8 @@ exec /bin/cp "$@"
 	c.Check(cmd.Calls(), HasLen, 0)
 }
 
-func mockLink(newLink func(string, string) error) (restore func()) {
-	oldLink := osLink
-	osLink = newLink
-	return func() {
-		osLink = oldLink
-	}
-}
-
 func noLink() func() {
-	return mockLink(func(string, string) error { return errors.New("no.") })
+	return squashfs.MockLink(func(string, string) error { return errors.New("no.") })
 }
 
 func (s *SquashfsTestSuite) TestInstallNotCopyTwice(c *C) {
@@ -184,7 +191,7 @@ func (s *SquashfsTestSuite) TestInstallSeedNoLink(c *C) {
 
 func (s *SquashfsTestSuite) TestPath(c *C) {
 	p := "/path/to/foo.snap"
-	snap := New("/path/to/foo.snap")
+	snap := squashfs.New("/path/to/foo.snap")
 	c.Assert(snap.Path(), Equals, p)
 }
 
@@ -207,29 +214,6 @@ func (s *SquashfsTestSuite) TestListDir(c *C) {
 	c.Check(fileNames[2], Equals, "foo-hook")
 }
 
-func alike(a, b os.FileInfo, c *C, comment CommentInterface) {
-	c.Check(a, NotNil, comment)
-	c.Check(b, NotNil, comment)
-	if a == nil || b == nil {
-		return
-	}
-
-	// the .Name() of the root will be different on non-squashfs things
-	_, asq := a.(*stat)
-	_, bsq := b.(*stat)
-	if !((asq && a.Name() == "/") || (bsq && b.Name() == "/")) {
-		c.Check(a.Name(), Equals, b.Name(), comment)
-	}
-
-	c.Check(a.Mode(), Equals, b.Mode(), comment)
-	if a.Mode().IsRegular() {
-		c.Check(a.Size(), Equals, b.Size(), comment)
-	}
-	am := a.ModTime().UTC().Truncate(time.Minute)
-	bm := b.ModTime().UTC().Truncate(time.Minute)
-	c.Check(am.Equal(bm), Equals, true, Commentf("%s != %s (%s)", am, bm, comment))
-}
-
 func (s *SquashfsTestSuite) TestWalk(c *C) {
 	sub := "."
 	snap := makeSnap(c, "name: foo", "")
@@ -277,18 +261,18 @@ func (s *SquashfsTestSuite) TestWalk(c *C) {
 	})
 
 	for k := range fpw {
-		alike(sqw[k], fpw[k], c, Commentf(k))
-		alike(sdw[k], fpw[k], c, Commentf(k))
+		squashfs.Alike(sqw[k], fpw[k], c, Commentf(k))
+		squashfs.Alike(sdw[k], fpw[k], c, Commentf(k))
 	}
 
 	for k := range sqw {
-		alike(fpw[k], sqw[k], c, Commentf(k))
-		alike(sdw[k], sqw[k], c, Commentf(k))
+		squashfs.Alike(fpw[k], sqw[k], c, Commentf(k))
+		squashfs.Alike(sdw[k], sqw[k], c, Commentf(k))
 	}
 
 	for k := range sdw {
-		alike(fpw[k], sdw[k], c, Commentf(k))
-		alike(sqw[k], sdw[k], c, Commentf(k))
+		squashfs.Alike(fpw[k], sdw[k], c, Commentf(k))
+		squashfs.Alike(sqw[k], sdw[k], c, Commentf(k))
 	}
 
 }
@@ -361,28 +345,206 @@ EOF
 }
 
 func (s *SquashfsTestSuite) TestBuild(c *C) {
+	// please keep TestBuildUsesExcludes in sync with this one so it makes sense.
 	buildDir := c.MkDir()
 	err := os.MkdirAll(filepath.Join(buildDir, "/random/dir"), 0755)
 	c.Assert(err, IsNil)
 	err = ioutil.WriteFile(filepath.Join(buildDir, "data.bin"), []byte("data"), 0644)
 	c.Assert(err, IsNil)
+	err = ioutil.WriteFile(filepath.Join(buildDir, "random", "data.bin"), []byte("more data"), 0644)
+	c.Assert(err, IsNil)
 
-	snap := New(filepath.Join(c.MkDir(), "foo.snap"))
-	err = snap.Build(buildDir)
+	snap := squashfs.New(filepath.Join(c.MkDir(), "foo.snap"))
+	err = snap.Build(buildDir, "app")
 	c.Assert(err, IsNil)
 
 	// unsquashfs writes a funny header like:
 	//     "Parallel unsquashfs: Using 1 processor"
 	//     "1 inodes (1 blocks) to write"
-	outputWithHeader, err := exec.Command("unsquashfs", "-n", "-l", snap.path).Output()
+	outputWithHeader, err := exec.Command("unsquashfs", "-n", "-l", snap.Path()).Output()
 	c.Assert(err, IsNil)
 	split := strings.Split(string(outputWithHeader), "\n")
 	output := strings.Join(split[3:], "\n")
-	c.Assert(string(output), Equals, `squashfs-root
+	c.Assert(string(output), Equals, `
+squashfs-root
 squashfs-root/data.bin
 squashfs-root/random
+squashfs-root/random/data.bin
 squashfs-root/random/dir
+`[1:]) // skip the first newline :-)
+}
+
+func (s *SquashfsTestSuite) TestBuildUsesExcludes(c *C) {
+	// please keep TestBuild in sync with this one so it makes sense.
+	buildDir := c.MkDir()
+	err := os.MkdirAll(filepath.Join(buildDir, "/random/dir"), 0755)
+	c.Assert(err, IsNil)
+	err = ioutil.WriteFile(filepath.Join(buildDir, "data.bin"), []byte("data"), 0644)
+	c.Assert(err, IsNil)
+	err = ioutil.WriteFile(filepath.Join(buildDir, "random", "data.bin"), []byte("more data"), 0644)
+	c.Assert(err, IsNil)
+
+	excludesFilename := filepath.Join(buildDir, ".snapignore")
+	err = ioutil.WriteFile(excludesFilename, []byte(`
+# ignore just one of the data.bin files we just added (the toplevel one)
+data.bin
+# also ignore ourselves
+.snapignore
+# oh and anything called "dir" anywhere
+... dir
+`), 0644)
+	c.Assert(err, IsNil)
+
+	snap := squashfs.New(filepath.Join(c.MkDir(), "foo.snap"))
+	err = snap.Build(buildDir, "app", excludesFilename)
+	c.Assert(err, IsNil)
+
+	outputWithHeader, err := exec.Command("unsquashfs", "-n", "-l", snap.Path()).Output()
+	c.Assert(err, IsNil)
+	split := strings.Split(string(outputWithHeader), "\n")
+	output := strings.Join(split[3:], "\n")
+	// compare with TestBuild
+	c.Assert(string(output), Equals, `
+squashfs-root
+squashfs-root/random
+squashfs-root/random/data.bin
+`[1:]) // skip the first newline :-)
+}
+
+func (s *SquashfsTestSuite) TestBuildSupportsMultipleExcludesWithOnlyOneWildcardsFlag(c *C) {
+	defer squashfs.MockFromCore(func(mountDir, cmd string, args ...string) (*exec.Cmd, error) {
+		c.Check(mountDir, Equals, dirs.SnapMountDir)
+		c.Check(cmd, Equals, "/usr/bin/mksquashfs")
+		return nil, errors.New("bzzt")
+	})()
+	mksq := testutil.MockCommand(c, "mksquashfs", "")
+	defer mksq.Restore()
+
+	snapPath := filepath.Join(c.MkDir(), "foo.snap")
+	snap := squashfs.New(snapPath)
+	err := snap.Build(c.MkDir(), "core", "exclude1", "exclude2", "exclude3")
+	c.Assert(err, IsNil)
+	calls := mksq.Calls()
+	c.Assert(calls, HasLen, 1)
+	c.Check(calls[0], DeepEquals, []string{
+		// the usual:
+		"mksquashfs", ".", snapPath, "-noappend", "-comp", "xz", "-no-fragments", "-no-progress",
+		// the interesting bits:
+		"-wildcards", "-ef", "exclude1", "-ef", "exclude2", "-ef", "exclude3",
+	})
+}
+
+func (s *SquashfsTestSuite) TestBuildUsesMksquashfsFromCoreIfAvailable(c *C) {
+	usedFromCore := false
+	defer squashfs.MockFromCore(func(mountDir, cmd string, args ...string) (*exec.Cmd, error) {
+		usedFromCore = true
+		c.Check(mountDir, Equals, dirs.SnapMountDir)
+		c.Check(cmd, Equals, "/usr/bin/mksquashfs")
+		return &exec.Cmd{Path: "/bin/true"}, nil
+	})()
+	mksq := testutil.MockCommand(c, "mksquashfs", "exit 1")
+	defer mksq.Restore()
+
+	buildDir := c.MkDir()
+
+	snap := squashfs.New(filepath.Join(c.MkDir(), "foo.snap"))
+	err := snap.Build(buildDir, "app")
+	c.Assert(err, IsNil)
+	c.Check(usedFromCore, Equals, true)
+	c.Check(mksq.Calls(), HasLen, 0)
+}
+
+func (s *SquashfsTestSuite) TestBuildUsesMksquashfsFromClassicIfCoreUnavailable(c *C) {
+	triedFromCore := false
+	defer squashfs.MockFromCore(func(mountDir, cmd string, args ...string) (*exec.Cmd, error) {
+		triedFromCore = true
+		c.Check(mountDir, Equals, dirs.SnapMountDir)
+		c.Check(cmd, Equals, "/usr/bin/mksquashfs")
+		return nil, errors.New("bzzt")
+	})()
+	mksq := testutil.MockCommand(c, "mksquashfs", "")
+	defer mksq.Restore()
+
+	buildDir := c.MkDir()
+
+	snap := squashfs.New(filepath.Join(c.MkDir(), "foo.snap"))
+	err := snap.Build(buildDir, "app")
+	c.Assert(err, IsNil)
+	c.Check(triedFromCore, Equals, true)
+	c.Check(mksq.Calls(), HasLen, 1)
+}
+
+func (s *SquashfsTestSuite) TestBuildFailsIfNoMksquashfs(c *C) {
+	triedFromCore := false
+	defer squashfs.MockFromCore(func(mountDir, cmd string, args ...string) (*exec.Cmd, error) {
+		triedFromCore = true
+		c.Check(mountDir, Equals, dirs.SnapMountDir)
+		c.Check(cmd, Equals, "/usr/bin/mksquashfs")
+		return nil, errors.New("bzzt")
+	})()
+	mksq := testutil.MockCommand(c, "mksquashfs", "exit 1")
+	defer mksq.Restore()
+
+	buildDir := c.MkDir()
+
+	snap := squashfs.New(filepath.Join(c.MkDir(), "foo.snap"))
+	err := snap.Build(buildDir, "app")
+	c.Assert(err, ErrorMatches, "mksquashfs call failed:.*")
+	c.Check(triedFromCore, Equals, true)
+	c.Check(mksq.Calls(), HasLen, 1)
+}
+
+func (s *SquashfsTestSuite) TestBuildVariesArgsByType(c *C) {
+	defer squashfs.MockFromCore(func(mountDir, cmd string, args ...string) (*exec.Cmd, error) {
+		c.Check(mountDir, Equals, dirs.SnapMountDir)
+		return nil, errors.New("bzzt")
+	})()
+	mksq := testutil.MockCommand(c, "mksquashfs", "")
+	defer mksq.Restore()
+
+	buildDir := c.MkDir()
+	filename := filepath.Join(c.MkDir(), "foo.snap")
+	snap := squashfs.New(filename)
+
+	permissiveTypeArgs := []string{".", filename, "-noappend", "-comp", "xz", "-no-fragments", "-no-progress"}
+	restrictedTypeArgs := append(permissiveTypeArgs, "-all-root", "-no-xattrs")
+	tests := []struct {
+		snapType string
+		args     []string
+	}{
+		{"", restrictedTypeArgs},
+		{"app", restrictedTypeArgs},
+		{"gadget", restrictedTypeArgs},
+		{"kernel", restrictedTypeArgs},
+		{"snapd", restrictedTypeArgs},
+		{"base", permissiveTypeArgs},
+		{"os", permissiveTypeArgs},
+		{"core", permissiveTypeArgs},
+	}
+
+	for _, t := range tests {
+		mksq.ForgetCalls()
+		comm := Commentf("type: %s", t.snapType)
+
+		c.Check(snap.Build(buildDir, t.snapType), IsNil, comm)
+		c.Assert(mksq.Calls(), HasLen, 1, comm)
+		c.Assert(mksq.Calls()[0], HasLen, len(t.args)+1)
+		c.Check(mksq.Calls()[0][0], Equals, "mksquashfs", comm)
+		c.Check(mksq.Calls()[0][1:], DeepEquals, t.args, comm)
+	}
+}
+
+func (s *SquashfsTestSuite) TestBuildReportsFailures(c *C) {
+	mockUnsquashfs := testutil.MockCommand(c, "mksquashfs", `
+echo Yeah, nah. >&2
+exit 1
 `)
+	defer mockUnsquashfs.Restore()
+
+	data := "mock kernel snap"
+	dir := makeSnapContents(c, "", data)
+	snap := squashfs.New("foo.snap")
+	c.Check(snap.Build(dir, "kernel"), ErrorMatches, `mksquashfs call failed: Yeah, nah.`)
 }
 
 func (s *SquashfsTestSuite) TestUnsquashfsStderrWriter(c *C) {
@@ -415,7 +577,7 @@ func (s *SquashfsTestSuite) TestUnsquashfsStderrWriter(c *C) {
 			expectedErr: `failed: "failed 1", "failed 2", "3 Failed", "4 Failed", and 1 more`,
 		},
 	} {
-		usw := newUnsquashfsStderrWriter()
+		usw := squashfs.NewUnsquashfsStderrWriter()
 		for _, l := range t.inp {
 			usw.Write([]byte(l))
 		}
@@ -436,9 +598,9 @@ func (s *SquashfsTestSuite) TestBuildDate(c *C) {
 	c.Assert(os.Chtimes(d, then, then), IsNil)
 	// make a snap using this directory
 	filename := filepath.Join(c.MkDir(), "foo.snap")
-	snap := New(filename)
-	c.Assert(snap.Build(d), IsNil)
+	snap := squashfs.New(filename)
+	c.Assert(snap.Build(d, "app"), IsNil)
 	// and see it's BuildDate is _now_, not _then_.
-	c.Check(BuildDate(filename), Equals, snap.BuildDate())
+	c.Check(squashfs.BuildDate(filename), Equals, snap.BuildDate())
 	c.Check(math.Abs(now.Sub(snap.BuildDate()).Seconds()) <= 61, Equals, true, Commentf("Unexpected build date %s", snap.BuildDate()))
 }
